from typing import List, Optional, TYPE_CHECKING, Union

from efootprint.abstract_modeling_classes.explainable_quantity import ExplainableQuantity
from efootprint.abstract_modeling_classes.empty_explainable_object import EmptyExplainableObject
from efootprint.constants.sources import Sources
from efootprint.abstract_modeling_classes.source_objects import SourceValue
from efootprint.constants.units import u
from efootprint.core.hardware.edge_hardware import EdgeHardware
from efootprint.core.hardware.edge_storage import EdgeStorage
from efootprint.core.hardware.hardware_base import InsufficientCapacityError

if TYPE_CHECKING:
    from efootprint.core.usage.edge_usage_journey import EdgeUsageJourney
    from efootprint.core.usage.edge_usage_pattern import EdgeUsagePattern
    from efootprint.core.usage.recurrent_edge_process import RecurrentEdgeProcess


class EdgeDevice(EdgeHardware):
    default_values = {
        "carbon_footprint_fabrication": SourceValue(60 * u.kg),
        "power": SourceValue(30 * u.W),
        "lifespan": SourceValue(4 * u.year),
        "idle_power": SourceValue(5 * u.W),
        "ram": SourceValue(8 * u.GB),
        "compute": SourceValue(4 * u.cpu_core),
        "power_usage_effectiveness": SourceValue(1.0 * u.dimensionless),
        "utilization_rate": SourceValue(0.8 * u.dimensionless),
        "base_ram_consumption": SourceValue(1 * u.GB),
        "base_compute_consumption": SourceValue(0.1 * u.cpu_core),
    }

    def __init__(self, name: str, carbon_footprint_fabrication: ExplainableQuantity,
                 power: ExplainableQuantity, lifespan: ExplainableQuantity, idle_power: ExplainableQuantity,
                 ram: ExplainableQuantity, compute: ExplainableQuantity,
                 power_usage_effectiveness: ExplainableQuantity,
                 utilization_rate: ExplainableQuantity, base_ram_consumption: ExplainableQuantity,
                 base_compute_consumption: ExplainableQuantity, storage: EdgeStorage):
        super().__init__(name, carbon_footprint_fabrication, power, lifespan)
        self.available_compute_per_instance = EmptyExplainableObject()
        self.available_ram_per_instance = EmptyExplainableObject()
        self.unitary_hourly_compute_need_over_full_timespan = EmptyExplainableObject()
        self.unitary_hourly_ram_need_over_full_timespan = EmptyExplainableObject()
        self.unitary_power_over_full_timespan = EmptyExplainableObject()

        self.idle_power = idle_power.set_label(f"Idle power of {self.name}")
        self.ram = ram.set_label(f"RAM of {self.name}")
        self.compute = compute.set_label(f"Compute of {self.name}")
        self.power_usage_effectiveness = power_usage_effectiveness.set_label(f"PUE of {self.name}")
        self.utilization_rate = utilization_rate.set_label(f"{self.name} utilization rate")
        self.base_ram_consumption = base_ram_consumption.set_label(f"Base RAM consumption of {self.name}")
        self.base_compute_consumption = base_compute_consumption.set_label(f"Base compute consumption of {self.name}")
        self.storage = storage

    @property
    def calculated_attributes(self):
        return [
            "available_ram_per_instance", "available_compute_per_instance",
            "unitary_hourly_ram_need_over_full_timespan", "unitary_hourly_compute_need_over_full_timespan",
            "unitary_power_over_full_timespan", "nb_of_instances", "instances_energy",
            "energy_footprint", "instances_fabrication_footprint"]

    @property
    def edge_usage_journey(self) -> Optional["EdgeUsageJourney"]:
        if self.modeling_obj_containers:
            if len(self.modeling_obj_containers) > 1:
                raise PermissionError(
                    f"EdgeDevice object can only be associated with one EdgeUsageJourney object but {self.name} is "
                    f"associated with {[mod_obj.name for mod_obj in self.modeling_obj_containers]}")
            return self.modeling_obj_containers[0]
        return None

    @property
    def edge_usage_pattern(self) -> Optional["EdgeUsagePattern"]:
        if self.modeling_obj_containers:
            return self.edge_usage_journey.edge_usage_pattern
        return None

    @property
    def modeling_objects_whose_attributes_depend_directly_on_me(self) -> List:
        return [self.storage]

    @property
    def edge_processes(self) -> List["RecurrentEdgeProcess"]:
        if self.modeling_obj_containers:
            return self.edge_usage_journey.edge_processes
        return []

    def update_available_ram_per_instance(self):
        available_ram_per_instance = (self.ram * self.utilization_rate - self.base_ram_consumption)
        
        if available_ram_per_instance.value < 0 * u.B:
            raise InsufficientCapacityError(
                self, "RAM", self.ram * self.utilization_rate, self.base_ram_consumption)

        self.available_ram_per_instance = available_ram_per_instance.set_label(
            f"Available RAM per {self.name} instance")

    def update_available_compute_per_instance(self):
        available_compute_per_instance = (self.compute * self.utilization_rate - self.base_compute_consumption)

        if available_compute_per_instance.value < 0 * u.cpu_core:
            raise InsufficientCapacityError(
                self, "compute", self.compute * self.utilization_rate, self.base_compute_consumption)

        self.available_compute_per_instance = available_compute_per_instance.set_label(
            f"Available compute per {self.name} instance")

    def update_unitary_hourly_ram_need_over_full_timespan(self):
        unitary_hourly_ram_need_over_full_timespan = sum(
            [edge_process.unitary_hourly_ram_need_over_full_timespan for edge_process in self.edge_processes],
            start=EmptyExplainableObject())

        max_ram_need = unitary_hourly_ram_need_over_full_timespan.max().to(u.GB)
        if max_ram_need > self.available_ram_per_instance:
            raise InsufficientCapacityError(
                self, "RAM", self.available_ram_per_instance, max_ram_need)

        self.unitary_hourly_ram_need_over_full_timespan = unitary_hourly_ram_need_over_full_timespan.to(u.GB).set_label(
            f"{self.name} hour by hour RAM need")

    def update_unitary_hourly_compute_need_over_full_timespan(self):
        unitary_hourly_compute_need_over_full_timespan = sum(
            [edge_process.unitary_hourly_compute_need_over_full_timespan for edge_process in self.edge_processes],
            start=EmptyExplainableObject())

        max_compute_need = unitary_hourly_compute_need_over_full_timespan.max().to(u.cpu_core)
        if max_compute_need > self.available_compute_per_instance:
            raise InsufficientCapacityError(
                self, "compute", self.available_compute_per_instance, max_compute_need)
        
        self.unitary_hourly_compute_need_over_full_timespan = unitary_hourly_compute_need_over_full_timespan.to(
            u.cpu_core).set_label(f"{self.name} hour by hour compute need")

    def update_unitary_power_over_full_timespan(self):
        unitary_compute_workload_over_full_timespan = (
                (self.unitary_hourly_compute_need_over_full_timespan + self.base_compute_consumption) / self.compute)

        unitary_power_over_full_timespan = (
                (self.idle_power + (self.power - self.idle_power) * unitary_compute_workload_over_full_timespan)
                * self.power_usage_effectiveness)

        self.unitary_power_over_full_timespan = unitary_power_over_full_timespan.set_label(
            f"{self.name} unitary power over full timespan.")



from abc import abstractmethod
from typing import List, TYPE_CHECKING, Union

from efootprint.abstract_modeling_classes.explainable_quantity import ExplainableQuantity
from efootprint.abstract_modeling_classes.empty_explainable_object import EmptyExplainableObject
from efootprint.abstract_modeling_classes.source_objects import SourceValue
from efootprint.constants.sources import Sources
from efootprint.constants.units import u
from efootprint.core.hardware.hardware_base import HardwareBase

if TYPE_CHECKING:
    from efootprint.core.usage.edge_usage_pattern import EdgeUsagePattern


class EdgeHardware(HardwareBase):
    def __init__(self, name: str, carbon_footprint_fabrication: ExplainableQuantity, power: ExplainableQuantity,
                 lifespan: ExplainableQuantity):
        super().__init__(
            name, carbon_footprint_fabrication, power, lifespan, SourceValue(1 * u.dimensionless))
        self.nb_of_instances = EmptyExplainableObject()
        self.instances_fabrication_footprint = EmptyExplainableObject()
        self.unitary_power_over_full_timespan = EmptyExplainableObject()
        self.instances_energy = EmptyExplainableObject()
        self.energy_footprint = EmptyExplainableObject()

    @property
    def modeling_objects_whose_attributes_depend_directly_on_me(self) -> List:
        return []

    @property
    def calculated_attributes(self):
        return ["nb_of_instances", "instances_fabrication_footprint",
                "unitary_power_over_full_timespan", "instances_energy", "energy_footprint"]

    @property
    def systems(self) -> List:
        return list(set(sum([elt.systems for elt in self.modeling_obj_containers], start=[])))

    @property
    @abstractmethod
    def edge_usage_pattern(self) -> "EdgeUsagePattern":
        pass

    @property
    def average_carbon_intensity(self) -> Union[ExplainableQuantity, EmptyExplainableObject]:
        edge_usage_pattern = self.edge_usage_pattern
        if edge_usage_pattern is not None:
            return edge_usage_pattern.country.average_carbon_intensity
        return EmptyExplainableObject()

    def update_nb_of_instances(self):
        if self.edge_usage_pattern:
            nb_of_instances = self.edge_usage_pattern.nb_edge_usage_journeys_in_parallel.copy()
        else:
            nb_of_instances = EmptyExplainableObject()

        self.nb_of_instances = nb_of_instances.set_label(f"{self.name} nb_of_instances")

    def update_instances_fabrication_footprint(self):
        instances_fabrication_footprint = (
                self.carbon_footprint_fabrication * self.nb_of_instances * ExplainableQuantity(1 * u.hour, "one hour")
                / self.lifespan)

        self.instances_fabrication_footprint = instances_fabrication_footprint.to(u.kg).set_label(
                f"Hourly {self.name} instances fabrication footprint")

    @abstractmethod
    def update_unitary_power_over_full_timespan(self):
        pass

    def update_instances_energy(self):
        unitary_energy_over_full_timespan = (
            self.unitary_power_over_full_timespan * ExplainableQuantity(1 * u.hour, "one hour")
        )

        instances_energy = unitary_energy_over_full_timespan * self.nb_of_instances

        self.instances_energy = instances_energy.to(u.kWh).set_label(
            f"Hourly energy consumed by {self.name} instances")

    def update_energy_footprint(self):
        if getattr(self, "average_carbon_intensity", None) is None:
            raise ValueError(
                f"Variable 'average_carbon_intensity' is not defined in object {self.name}."
                f" This shouldn’t happen as server objects have it as input parameter and Storage as property")
        energy_footprint = (self.instances_energy * self.average_carbon_intensity)

        self.energy_footprint = energy_footprint.to(u.kg).set_label(f"Hourly {self.name} energy footprint")



from typing import List, TYPE_CHECKING, Optional
import numpy as np
from pint import Quantity

from efootprint.abstract_modeling_classes.empty_explainable_object import EmptyExplainableObject
from efootprint.abstract_modeling_classes.explainable_recurring_quantities import ExplainableRecurringQuantities
from efootprint.abstract_modeling_classes.modeling_object import ModelingObject
from efootprint.abstract_modeling_classes.source_objects import SourceRecurringValues
from efootprint.constants.units import u

if TYPE_CHECKING:
    from efootprint.core.usage.edge_usage_pattern import EdgeUsagePattern
    from efootprint.core.usage.edge_usage_journey import EdgeUsageJourney
    from efootprint.core.hardware.edge_device import EdgeDevice
    from efootprint.core.system import System


class RecurrentEdgeProcess(ModelingObject):
    default_values = {
        "recurrent_compute_needed": SourceRecurringValues(Quantity(np.array([1] * 168, dtype=np.float32), u.cpu_core)),
        "recurrent_ram_needed": SourceRecurringValues(Quantity(np.array([1] * 168, dtype=np.float32), u.GB)),
        "recurrent_storage_needed": SourceRecurringValues(Quantity(np.array([0] * 168, dtype=np.float32), u.GB)),
    }

    def __init__(self, name: str, recurrent_compute_needed: ExplainableRecurringQuantities, 
                 recurrent_ram_needed: ExplainableRecurringQuantities,
                 recurrent_storage_needed: ExplainableRecurringQuantities):
        super().__init__(name)
        self.unitary_hourly_compute_need_over_full_timespan = EmptyExplainableObject()
        self.unitary_hourly_ram_need_over_full_timespan = EmptyExplainableObject()
        self.unitary_hourly_storage_need_over_full_timespan = EmptyExplainableObject()
        
        self.recurrent_compute_needed = recurrent_compute_needed.set_label(f"{self.name} recurrent compute needed")
        self.recurrent_ram_needed = recurrent_ram_needed.set_label(f"{self.name} recurrent ram needed")
        self.recurrent_storage_needed = recurrent_storage_needed.set_label(f"{self.name} recurrent storage needed")

    @property
    def calculated_attributes(self):
        return ["unitary_hourly_compute_need_over_full_timespan", "unitary_hourly_ram_need_over_full_timespan",
                "unitary_hourly_storage_need_over_full_timespan"]

    @property
    def edge_usage_journey(self) -> Optional["EdgeUsageJourney"]:
        if self.modeling_obj_containers:
            if len(self.modeling_obj_containers) > 1:
                raise PermissionError(
                    f"RecurrentEdgeProcess object can only be associated with one EdgeUsageJourney object but {self.name} is "
                    f"associated with {[mod_obj.name for mod_obj in self.modeling_obj_containers]}")
            return self.modeling_obj_containers[0]
        else:
            return None

    @property
    def edge_usage_pattern(self) -> Optional["EdgeUsagePattern"]:
        if self.modeling_obj_containers:
            return self.edge_usage_journey.edge_usage_pattern
        else:
            return None

    @property
    def edge_device(self) -> Optional["EdgeDevice"]:
        if self.modeling_obj_containers:
            return self.edge_usage_journey.edge_device
        else:
            return None
        
    @property
    def systems(self) -> List["System"]:
        if self.modeling_obj_containers:
            return self.edge_usage_journey.systems
        return []

    @property
    def modeling_objects_whose_attributes_depend_directly_on_me(self) -> List:
        return []

    def update_unitary_hourly_compute_need_over_full_timespan(self):
        if self.edge_usage_pattern is None:
            self.unitary_hourly_compute_need_over_full_timespan = EmptyExplainableObject()
            return
        self.unitary_hourly_compute_need_over_full_timespan = (
            self.recurrent_compute_needed.generate_hourly_quantities_over_timespan(
                self.edge_usage_pattern.hourly_edge_usage_journey_starts, 
                self.edge_usage_pattern.country.timezone))

    def update_unitary_hourly_ram_need_over_full_timespan(self):
        if self.edge_usage_pattern is None:
            self.unitary_hourly_ram_need_over_full_timespan = EmptyExplainableObject()
            return
        self.unitary_hourly_ram_need_over_full_timespan = (
            self.recurrent_ram_needed.generate_hourly_quantities_over_timespan(
                self.edge_usage_pattern.hourly_edge_usage_journey_starts, 
                self.edge_usage_pattern.country.timezone))

    def update_unitary_hourly_storage_need_over_full_timespan(self):
        if self.edge_usage_pattern is None:
            self.unitary_hourly_storage_need_over_full_timespan = EmptyExplainableObject()
            return
        self.unitary_hourly_storage_need_over_full_timespan = (
            self.recurrent_storage_needed.generate_hourly_quantities_over_timespan(
                self.edge_usage_pattern.hourly_edge_usage_journey_starts,
                self.edge_usage_pattern.country.timezone))



from typing import List, TYPE_CHECKING, Optional

from efootprint.abstract_modeling_classes.explainable_quantity import ExplainableQuantity
from efootprint.abstract_modeling_classes.modeling_object import ModelingObject
from efootprint.abstract_modeling_classes.source_objects import SourceValue
from efootprint.constants.sources import Sources
from efootprint.constants.units import u
from efootprint.core.hardware.hardware_base import InsufficientCapacityError
from efootprint.core.usage.recurrent_edge_process import RecurrentEdgeProcess
from efootprint.core.hardware.edge_device import EdgeDevice

if TYPE_CHECKING:
    from efootprint.core.usage.edge_usage_pattern import EdgeUsagePattern
    from efootprint.core.system import System


class EdgeUsageJourney(ModelingObject):
    default_values = {
        "usage_span": SourceValue(6 * u.year)
    }

    def __init__(self, name: str, edge_processes: List[RecurrentEdgeProcess], edge_device: EdgeDevice,
                 usage_span: ExplainableQuantity):
        super().__init__(name)
        self.assert_usage_span_is_inferior_to_edge_device_lifespan(usage_span, edge_device)
        self.edge_processes = edge_processes
        self.edge_device = edge_device
        self.usage_span = usage_span.set_label(f"Usage span of {self.name}")

    @staticmethod
    def assert_usage_span_is_inferior_to_edge_device_lifespan(usage_span: ExplainableQuantity, edge_device: EdgeDevice):
        if usage_span > edge_device.lifespan:
            raise InsufficientCapacityError(edge_device, "lifespan", edge_device.lifespan, usage_span)

    @property
    def edge_usage_pattern(self) -> Optional["EdgeUsagePattern"]:
        if self.modeling_obj_containers:
            if len(self.modeling_obj_containers) > 1:
                raise PermissionError(
                    f"EdgeUsageJourney object can only be associated with one EdgeUsagePattern object but {self.name} "
                    f"is associated with {[mod_obj.name for mod_obj in self.modeling_obj_containers]}")
            return self.modeling_obj_containers[0]
        else:
            return None

    @property
    def systems(self) -> List["System"]:
        if self.modeling_obj_containers:
            return self.edge_usage_pattern.systems
        return []

    @property
    def modeling_objects_whose_attributes_depend_directly_on_me(self) -> List["EdgeUsagePattern"] | List[RecurrentEdgeProcess]:
        if self.edge_usage_pattern:
            return [self.edge_usage_pattern]
        return self.edge_processes + [self.edge_device]

    def __setattr__(self, name, input_value, check_input_validity=True):
        if name == "usage_span":
            self.assert_usage_span_is_inferior_to_edge_device_lifespan(input_value, self.edge_device)
        super().__setattr__(name, input_value, check_input_validity=check_input_validity)
